#ifndef MSP_DATAFILE_COLLECTION_H_
#define MSP_DATAFILE_COLLECTION_H_

#include <msp/core/maputils.h>
#include <msp/core/meta.h>
#include <msp/core/refptr.h>
#include "collectionsource.h"
#include "loader.h"

/* XXX This file is a big mess with too many things in it.  However, the
dependencies between those things make it difficult to split up. */

namespace Msp {
namespace DataFile {

/**
Helper struct to determine whether a Loader has a Collection typedef.
*/
template<typename T>
struct NeedsCollection: public Sfinae
{
	template<typename U>
	static Yes f(typename U::Collection *);
	template<typename U>
	static No f(...);

	enum { value = Evaluate<sizeof(f<T>(0))>::value };
};

class CollectionItemTypeBase;

template<typename T>
class CollectionItemType;

/**
A collection of objects that can be loaded from a datafile.  Each object is
identified by a name, which must be unique across the entire collection.

While this class can be instantiated by itself and used for storing objects,
loading requires that a subclass defines the supported types.  See the add_type
method for details.

Collections can have sources for loading objects on demand.  Automatic loading
only works on a non-const Collection.  See class CollectionSource for details.

As a last resort, a fallback collection can be designated for loading items
that are not present.  Items retrieted from the fallback collection are shared
between the collections, and are only deleted when all collections in the chain
have been destroyed.
*/
class Collection
{
public:
	/**
	Loads objects into a Collection.  Automatically picks up keywords from
	defined item types.
	*/
	class Loader: public DataFile::Loader
	{
		template<typename T> friend class CollectionItemType;

	private:
		Collection &coll;

	public:
		Loader(Collection &);
		Collection &get_object() const { return coll; }
	private:
		template<typename T, typename S>
		void item(const std::string &n)
		{
			RefPtr<T> it = new T;
			ItemLoader<T> ldr(*it, coll);
			load_sub_with(ldr);
			coll.add<S>(n, it.get());
			it.release();
		}
	};

	template<typename T, bool = NeedsCollection<typename T::Loader>::value>
	class ItemLoader;

private:
	typedef std::map<std::string, Variant> ItemMap;
	typedef std::list<CollectionItemTypeBase *> TypeList;
	typedef std::list<CollectionSource *> SourceList;

	TypeList types;
	ItemMap items;
	SourceList sources;
	Collection *fallback;

	Collection(const Collection &);
	Collection &operator=(const Collection &);
public:
	Collection();
	virtual ~Collection();

	/** Adds an object into the collection.  The name must not pre-exist.  The
	collection takes ownership of the object. */
	template<typename T>
	void add(const std::string &name, T *item)
	{
		if(!item)
			throw std::invalid_argument("Collection::add(item)");

		RefPtr<typename RemoveConst<T>::Type> ptr(item);
		try
		{
			insert_unique(items, name, ptr);
		}
		catch(...)
		{
			// Avoid deleting the object
			ptr.release();
			throw;
		}
	}

	/// Gets a typed object from the collection.
	template<typename T>
	T &get(const std::string &name) const
	{
		return extract<typename RemoveConst<T>::Type>(get_item(items, name));
	}

	/** Gets a typed object from the collection.  If the name is not found,
	automatic creation with the type's creator function (if defined) or from
	sources (if present) is attempted. */
	template<typename T>
	T &get(const std::string &name)
	{
		typedef typename RemoveConst<T>::Type NCT;
		return extract<NCT>(get_var(name, get_type<NCT>()));
	}

private:
	const Variant &get_var(const std::string &, const CollectionItemTypeBase *);

	template<typename T>
	T &extract(const Variant &var) const;

	template<typename T>
	std::list<T *> extract_list(const std::list<const Variant *> &vars) const
	{
		std::list<T *> result;
		for(std::list<const Variant *>::const_iterator i=vars.begin(); i!=vars.end(); ++i)
			result.push_back(&extract<T>(**i));
		return result;
	}

	void gather_items(std::list<const Variant *> *, std::list<std::string> *, const CollectionItemTypeBase &, bool) const;

	template<typename T>
	void gather_items(std::list<const Variant *> *vars, std::list<std::string> *names, const CollectionItemTypeBase *type, bool include_sources) const
	{
		if(type || (type = get_type<T>()))
			gather_items(vars, names, *type, include_sources);
		else
			gather_items(vars, names, CollectionItemType<T>(), false);
	}

public:
	/// Returns a list of the names of objects of one type in the collection.
	template<typename T>
	std::list<std::string> get_names() const
	{
		std::list<std::string> names;
		gather_items<typename RemoveConst<T>::Type>(0, &names, 0, false);
		return names;
	}

	/** Returns a list of the names of objects of one type in the collection or
	available from sources. */
	template<typename T>
	std::list<std::string> get_names()
	{
		std::list<std::string> names;
		gather_items<typename RemoveConst<T>::Type>(0, &names, 0, true);
		return names;
	}

	/// Returns a list of objects of one type in the collection.
	template<typename T>
	std::list<T *> get_list() const
	{
		std::list<const Variant *> vars;
		gather_items<typename RemoveConst<T>::Type>(&vars, 0, 0, false);
		return extract_list<T>(vars);
	}

	/** Returns a list of objects of one type, loading them from sources if
	necessary. */
	template<typename T>
	std::list<T *> get_list()
	{
		CollectionItemTypeBase *type = get_type<typename RemoveConst<T>::Type>();
		if(type)
			load_items_from_sources(*type);

		std::list<const Variant *> vars;
		gather_items<typename RemoveConst<T>::Type>(&vars, 0, type, true);
		return extract_list<T>(vars);
	}

private:
	unsigned get_status(const std::string &, const CollectionItemTypeBase &) const;

	template<typename T>
	unsigned get_status(const std::string &name) const
	{
		// XXX Should go through all applicable types
		if(CollectionItemTypeBase *type = get_type<T>())
			return get_status(name, *type);

		ItemMap::const_iterator i = items.find(name);
		return (i!=items.end() && i->second.check_type<RefPtr<T> >());
	}

public:
	/// Checks whether a typed object exists in the collection.
	template<typename T>
	bool contains(const std::string &name) const
	{ return get_status<typename RemoveConst<T>::Type>(name)==1; }

	/** Checks whether a typed object exists in the collection or is loadable
	from a source. */
	template<typename T>
	bool contains(const std::string &name)
	{ return get_status<typename RemoveConst<T>::Type>(name)>0; }

	/// Returns the name of an item in the collection.
	template<typename T>
	const std::string &get_name(T *d) const
	{
		typedef RefPtr<typename RemoveConst<T>::Type> RPNCT;

		for(ItemMap::const_iterator i=items.begin(); i!=items.end(); ++i)
			if(i->second.check_type<RPNCT>())
				if(i->second.value<RPNCT>().get()==d)
					return i->first;
	
		// XXX Need better exception class
		throw std::runtime_error("Item not found in collection");
	}

protected:
	/** Adds a type to the collection.  The returned descriptor object reference
	can be used to define how objects of that type can be loaded. */
	template<typename T>
	CollectionItemType<T> &add_type();

private:
	/// Returns the descriptor for a type, or null if one isn't defined.
	template<typename T>
	CollectionItemTypeBase *get_type() const;

	/// Returns the descriptor for an item, or null if it's of an unknown type.
	CollectionItemTypeBase *get_type_for_item(const Variant &) const;

protected:
	/** Adds a source for automatically loading items.  Sources are consulted
	in the order they are added. */
	void add_source(CollectionSource &);

public:
	/** Opens a raw resource, without interpreting it as object data.  Null is
	returned if no such file is found.  The caller must dispose of the returned
	object when done with it. */
	IO::Seekable *open_raw(const std::string &) const;

protected:
	IO::Seekable *open_from_sources(const std::string &n) { return open_raw(n); }

private:
	void gather_names_from_sources(std::list<std::string> &, const CollectionItemTypeBase &) const;

	void load_items_from_sources(const CollectionItemTypeBase &);

protected:
	/** Sets a fallback collection, which will be consulted if an item is not
	found. */
	void set_fallback(Collection *);
};

template<typename T>
class Collection::ItemLoader<T, false>: public T::Loader
{
public:
	ItemLoader(T &o, Collection &):
		T::Loader(o)
	{ }
};

template<typename T>
class Collection::ItemLoader<T, true>: public T::Loader
{
public:
	ItemLoader(T &o, Collection &c):
		T::Loader(o, dynamic_cast<typename T::Loader::Collection &>(c))
	{ }
};


class CollectionItemTypeBase
{
protected:
	struct ExtractorBase
	{
		virtual ~ExtractorBase() { }
	};

	template<typename T>
	struct Extractor: ExtractorBase
	{
		virtual T &extract(const Variant &) const = 0;
	};

	std::string kwd;
	std::vector<std::string> suffixes;
	std::vector<ExtractorBase *> extractors;

	CollectionItemTypeBase() { }
public:
	virtual ~CollectionItemTypeBase();

	void set_keyword(const std::string &);
	const std::string &get_keyword() const { return kwd; }
	void add_suffix(const std::string &);
	bool match_name(const std::string &) const;
	virtual bool is_same_type(const CollectionItemTypeBase &) const = 0;
	virtual bool check_item_type(const Variant &) const = 0;
	virtual void add_to_loader(Collection::Loader &) const = 0;
	virtual bool can_create() const = 0;
	virtual void create_item(Collection &, const std::string &) const = 0;
	virtual void load_item(Collection &, Parser &, const std::string &) const = 0;

	template<typename T>
	bool can_extract() const
	{
		for(std::vector<ExtractorBase *>::const_iterator i=extractors.begin(); i!=extractors.end(); ++i)
			if(dynamic_cast<Extractor<T> *>(*i))
				return true;
		return false;
	}

	template<typename T>
	T *extract(const Variant &var) const
	{
		for(std::vector<ExtractorBase *>::const_iterator i=extractors.begin(); i!=extractors.end(); ++i)
			if(Extractor<T> *ex = dynamic_cast<Extractor<T> *>(*i))
				return &ex->extract(var);
		return 0;
	}
};


/**
Describes a type of item that can be loaded by a Collection.  These are created
by Collection::add_type.
*/
template<typename T>
class CollectionItemType: public CollectionItemTypeBase
{
private:
	struct CreatorBase
	{
		virtual ~CreatorBase() { }

		virtual T *create(Collection &, const std::string &) const = 0;
	};

	template<typename C>
	struct Creator: CreatorBase
	{
		typedef T *(C::*FuncPtr)(const std::string &);

		FuncPtr func;

		Creator(FuncPtr f): func(f) { }

		virtual T *create(Collection &coll, const std::string &name) const
		{ return (static_cast<C &>(coll).*func)(name); }
	};

	template<typename B>
	struct Extractor: CollectionItemTypeBase::Extractor<B>
	{
		virtual B &extract(const Variant &var) const
		{ return *var.value<RefPtr<T> >(); }
	};

	CreatorBase *creat;

public:
	CollectionItemType():
		creat(0)
	{ }

	~CollectionItemType()
	{
		delete creat;
	}

	/** Sets a datafile keyword for this item type.  The Collection's loader
	will accept a statement with this keyword and a single string argument - the
	item's name. */
	CollectionItemType &keyword(const std::string &k)
	{
		set_keyword(k);
		return *this;
	}

	/** Adds a suffix that is used to match names when looking for future
	objects.  There is no implied separator; a name matches if it ends with the
	suffix.  If a keyword is defined before any suffixes, then "."+keyword is
	added as a suffix. */
	CollectionItemType &suffix(const std::string &s)
	{
		add_suffix(s);
		return *this;
	}

	/** Attaches a creator function to this item type.  If an item is not found
	in the Collection, the creator function for its type is called to create it.
	The function must be a member of the Collection subclass containing the
	type.  It must return the created object, or null if it could not be
	created.  It's also permissible to load the item via other means and then
	return null. */
	template<typename C>
	CollectionItemType &creator(T *(C::*func)(const std::string &))
	{
		delete creat;
		creat = new Creator<C>(func);
		return *this;
	}

	/** Makes items of this type available through a base class. */
	template<typename B>
	CollectionItemType &base()
	{
		extractors.push_back(new Extractor<B>);
		return *this;
	}

	virtual bool is_same_type(const CollectionItemTypeBase &other) const
	{ return dynamic_cast<const CollectionItemType<T> *>(&other); }

	virtual bool check_item_type(const Variant &var) const
	{ return var.check_type<RefPtr<T> >(); }

	virtual void add_to_loader(Collection::Loader &loader) const
	{ loader.add(kwd, &Collection::Loader::item<T, T>); }

	virtual bool can_create() const
	{ return creat!=0; }

	virtual void create_item(Collection &coll, const std::string &name) const
	{
		if(!creat)
			throw std::runtime_error("no creator");
		T *obj = creat->create(coll, name);
		if(obj)
			coll.add(name, obj);
	}

	virtual void load_item(Collection &coll, Parser &parser, const std::string &name) const
	{
		RefPtr<T> obj = new T;
		Collection::ItemLoader<T> ldr(*obj, coll);
		ldr.load(parser);
		coll.add(name, obj.get());
		obj.release();
	}
};


template<typename T>
T &Collection::extract(const Variant &var) const
{
	if(!var.check_type<RefPtr<T> >())
		if(CollectionItemTypeBase *type = get_type_for_item(var))
			if(T *item = type->extract<T>(var))
				return *item;

	return *var.value<RefPtr<T> >();
}

template<typename T>
CollectionItemType<T> &Collection::add_type()
{
	CollectionItemType<T> *type = new CollectionItemType<T>;
	types.push_back(type);
	return *type;
}

template<typename T>
CollectionItemTypeBase *Collection::get_type() const
{
	for(TypeList::const_iterator j=types.begin(); j!=types.end(); ++j)
		if(dynamic_cast<CollectionItemType<T> *>(*j))
			return *j;
	for(TypeList::const_iterator j=types.begin(); j!=types.end(); ++j)
		if((*j)->can_extract<T>())
			return *j;
	return 0;
}

} // namespace DataFile
} // namespace Msp

#endif
