#ifndef MSP_DATAFILE_LOADERACTION_H_
#define MSP_DATAFILE_LOADERACTION_H_

#include "argumentstore.h"
#include "statement.h"

namespace Msp {
namespace DataFile {

class Loader;

/**
Base class for loader actions.
*/
class LoaderAction
{
protected:
	LoaderAction() { }
public:
	virtual ~LoaderAction() { }

	/** Called to process a statement. */
	virtual void execute(Loader &, const Statement &) const = 0;

	virtual void execute(Loader &, const ArgumentStore &) const = 0;

	virtual std::string get_signature() const = 0;
};


/**
Loads a statement by calling a function that takes no arguments.
*/
template<typename L>
class LoaderFunc0: public LoaderAction
{
private:
	typedef void (L::*FuncType)();

	FuncType func;

public:
	LoaderFunc0(FuncType f): func(f) { }

	virtual void execute(Loader &l, const Statement &) const
	{
		(dynamic_cast<L &>(l).*func)();
	};

	virtual void execute(Loader &l, const ArgumentStore &) const
	{
		(dynamic_cast<L &>(l).*func)();
	};

	virtual std::string get_signature() const
	{ return std::string(); }
};


/**
Loads a statement by calling a function that takes one argument.
*/
template<typename L, typename A0>
class LoaderFunc1: public LoaderAction
{
private:
	typedef void (L::*FuncType)(A0);

	FuncType func;

public:
	LoaderFunc1(FuncType f): func(f) { }

	virtual void execute(Loader &l, const Statement &st) const
	{
		(dynamic_cast<L &>(l).*func)(st.args[0].get<A0>());
	}

	virtual void execute(Loader &l, const ArgumentStore &as) const
	{
		(dynamic_cast<L &>(l).*func)(as.get<A0>(0));
	}

	virtual std::string get_signature() const
	{ return std::string(1, TypeInfo<A0>::signature); }
};


/**
Loads a statement by calling a function that takes an array of values.
*/
template<typename L, typename A0>
class LoaderFunc1<L, const std::vector<A0> &>: public LoaderAction
{
private:
	typedef void (L::*FuncType)(const std::vector<A0> &);

	FuncType func;

public:
	LoaderFunc1(FuncType f): func(f) { }

	virtual void execute(Loader &l, const Statement &st) const
	{
		std::vector<A0> values;
		values.reserve(st.args.size());
		for(Statement::Arguments::const_iterator i=st.args.begin(); i!=st.args.end(); ++i)
			values.push_back(i->get<A0>());
		(dynamic_cast<L &>(l).*func)(values);
	}

	virtual void execute(Loader &l, const ArgumentStore &as) const
	{
		std::vector<A0> values;
		unsigned n_args = as.get_info().key.signature.size();
		values.reserve(n_args);
		for(unsigned i=0; i<n_args; ++i)
			values.push_back(as.get<A0>(i));
		(dynamic_cast<L &>(l).*func)(values);
	}

	virtual std::string get_signature() const
	{
		std::string result;
		result += TypeInfo<A0>::signature;
		result += '*';
		return result;
	}
};


/**
Loads a statement by calling a function with the statement itself as argument.
*/
template<typename L>
class LoaderFunc1<L, const Statement &>: public LoaderAction
{
private:
	typedef void (L::*FuncType)(const Statement &);

	FuncType func;

public:
	LoaderFunc1(FuncType f): func(f) { }

	virtual void execute(Loader &l, const Statement &st) const
	{
		(dynamic_cast<L &>(l).*func)(st);
	}

	virtual void execute(Loader &, const ArgumentStore &) const
	{
		throw std::logic_error("incompatible format");
	}

	virtual std::string get_signature() const
	{ return "*"; }
};


template<typename L, typename A0, typename A1>
class LoaderFunc2: public LoaderAction
{
private:
	typedef void (L::*FuncType)(A0, A1);

	FuncType func;

public:
	LoaderFunc2(FuncType f): func(f) { }

	virtual void execute(Loader &l, const Statement &st) const
	{
		(dynamic_cast<L &>(l).*func)(st.args[0].get<A0>(), st.args[1].get<A1>());
	}

	virtual void execute(Loader &l, const ArgumentStore &as) const
	{
		(dynamic_cast<L &>(l).*func)(as.get<A0>(0), as.get<A1>(1));
	}

	virtual std::string get_signature() const
	{
		std::string result;
		result += TypeInfo<A0>::signature;
		result += TypeInfo<A1>::signature;
		return result;
	}
};


template<typename L, typename A0, typename A1, typename A2>
class LoaderFunc3: public LoaderAction
{
private:
	typedef void (L::*FuncType)(A0, A1, A2);

	FuncType func;

public:
	LoaderFunc3(FuncType f): func(f) { }

	virtual void execute(Loader &l, const Statement &st) const
	{
		(dynamic_cast<L &>(l).*func)(st.args[0].get<A0>(), st.args[1].get<A1>(), st.args[2].get<A2>());
	}

	virtual void execute(Loader &l, const ArgumentStore &as) const
	{
		(dynamic_cast<L &>(l).*func)(as.get<A0>(0), as.get<A1>(1), as.get<A2>(2));
	}

	virtual std::string get_signature() const
	{
		std::string result;
		result += TypeInfo<A0>::signature;
		result += TypeInfo<A1>::signature;
		result += TypeInfo<A2>::signature;
		return result;
	}
};


template<typename L, typename A0, typename A1, typename A2, typename A3>
class LoaderFunc4: public LoaderAction
{
private:
	typedef void (L::*FuncType)(A0, A1, A2, A3);

	FuncType func;

public:
	LoaderFunc4(FuncType f): func(f) { }

	virtual void execute(Loader &l, const Statement &st) const
	{
		(dynamic_cast<L &>(l).*func)(st.args[0].get<A0>(), st.args[1].get<A1>(), st.args[2].get<A2>(), st.args[3].get<A3>());
	}

	virtual void execute(Loader &l, const ArgumentStore &as) const
	{
		(dynamic_cast<L &>(l).*func)(as.get<A0>(0), as.get<A1>(1), as.get<A2>(2), as.get<A3>(3));
	}

	virtual std::string get_signature() const
	{
		std::string result;
		result += TypeInfo<A0>::signature;
		result += TypeInfo<A1>::signature;
		result += TypeInfo<A2>::signature;
		result += TypeInfo<A3>::signature;
		return result;
	}
};


template<typename L, typename A0, typename A1, typename A2, typename A3, typename A4>
class LoaderFunc5: public LoaderAction
{
private:
	typedef void (L::*FuncType)(A0, A1, A2, A3, A4);

	FuncType func;

public:
	LoaderFunc5(FuncType f): func(f) { }

	virtual void execute(Loader &l, const Statement &st) const
	{
		(dynamic_cast<L &>(l).*func)(st.args[0].get<A0>(), st.args[1].get<A1>(), st.args[2].get<A2>(), st.args[3].get<A3>(), st.args[4].get<A4>());
	}

	virtual void execute(Loader &l, const ArgumentStore &as) const
	{
		(dynamic_cast<L &>(l).*func)(as.get<A0>(0), as.get<A1>(1), as.get<A2>(2), as.get<A3>(3), as.get<A4>(4));
	}

	virtual std::string get_signature() const
	{
		std::string result;
		result += TypeInfo<A0>::signature;
		result += TypeInfo<A1>::signature;
		result += TypeInfo<A2>::signature;
		result += TypeInfo<A3>::signature;
		result += TypeInfo<A4>::signature;
		return result;
	}
};


template<typename L, typename T0>
class LoadValue1: public LoaderAction
{
private:
	typedef T0 L::*Pointer0Type;

	Pointer0Type ptr0;

public:
	LoadValue1(Pointer0Type p0): ptr0(p0) { }

	virtual void execute(Loader &l, const Statement &st) const
	{
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr0 = st.args[0].get<T0>();
	}

	virtual void execute(Loader &l, const ArgumentStore &as) const
	{
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr0 = as.get<T0>(0);
	}

	virtual std::string get_signature() const
	{ return std::string(1, TypeInfo<T0>::signature); }
};


template<typename L, typename T0>
class LoadValue1<L, T0 *>: public LoaderAction
{
private:
	typedef T0 *L::*Pointer0Type;

	Pointer0Type ptr0;

public:
	LoadValue1(Pointer0Type p0): ptr0(p0) { }

	virtual void execute(Loader &l, const Statement &st) const
	{
		typename L::Loader &ldr = dynamic_cast<typename L::Loader &>(l);
		ldr.get_object().*ptr0 = &ldr.get_collection().template get<T0>(st.args[0].get<std::string>());
	}

	virtual void execute(Loader &l, const ArgumentStore &as) const
	{
		typename L::Loader &ldr = dynamic_cast<typename L::Loader &>(l);
		ldr.get_object().*ptr0 = &ldr.get_collection().template get<T0>(as.get<std::string>(0));
	}

	virtual std::string get_signature() const
	{ return std::string(1, TypeInfo<std::string>::signature); }
};


template<typename L, typename T0, typename T1>
class LoadValue2: public LoaderAction
{
private:
	typedef T0 L::*Pointer0Type;
	typedef T1 L::*Pointer1Type;

	Pointer0Type ptr0;
	Pointer1Type ptr1;

public:
	LoadValue2(Pointer0Type p0, Pointer1Type p1): ptr0(p0), ptr1(p1) { }

	virtual void execute(Loader &l, const Statement &st) const
	{
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr0 = st.args[0].get<T0>();
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr1 = st.args[1].get<T1>();
	}

	virtual void execute(Loader &l, const ArgumentStore &as) const
	{
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr0 = as.get<T0>(0);
		dynamic_cast<typename L::Loader &>(l).get_object().*ptr1 = as.get<T1>(1);
	}

	virtual std::string get_signature() const
	{
		std::string result;
		result += TypeInfo<T0>::signature;
		result += TypeInfo<T1>::signature;
		return result;
	}
};

} // namespace DataFile
} // namespace Msp

#endif
